-- This block arbitrates between multiple incoming avalon-mm transactions in a many to one fashion
-- It operates in a round robin format
-- It will wait for the current transaction to complete before sampling the next interface
-- It operates on a single clock domain
-- This operates with the "basic" port sizes used in my applications (see pkg for more details)
-- While the burstcount port exists, it currently only accepts burst_size 1 transactions
-- It can handle a variable amount of inputs, based on a generic

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.math_pkg.all;
use work.avalon_mm_pkg.all;

entity avalon_mm_basic_register_arbiter is
generic
(
  INPUTS : natural := 2
);
port
(
  clk : in  std_logic;
  rst : in  std_logic;

  address_array         : in  BASIC_REGISTER_ADDRESS_ARRAY(INPUTS - 1     downto 0);
  burst_count_array     : in  BASIC_REGISTER_BURST_COUNT_ARRAY(INPUTS - 1 downto 0);
  read_array            : in  std_logic_vector(INPUTS - 1                 downto 0);
  write_array           : in  std_logic_vector(INPUTS - 1                 downto 0);
  write_data_array      : in  BASIC_REGISTER_DATA_ARRAY(INPUTS - 1        downto 0);
  read_data_array       : out BASIC_REGISTER_DATA_ARRAY(INPUTS - 1        downto 0);
  read_data_valid_array : out std_logic_vector(INPUTS - 1                 downto 0);
  wait_request_array    : out std_logic_vector(INPUTS - 1                 downto 0);

  address         : out std_logic_vector(BASIC_REGISTER_ADDRESS_WIDTH - 1     downto 0);
  burst_count     : out std_logic_vector(BASIC_REGISTER_BURST_COUNT_WIDTH - 1 downto 0);
  read            : out std_logic;
  write           : out std_logic;
  write_data      : out std_logic_vector(BASIC_REGISTER_DATA_WIDTH - 1        downto 0);
  read_data       : in  std_logic_vector(BASIC_REGISTER_DATA_WIDTH - 1        downto 0);
  read_data_valid : in  std_logic;
  wait_request    : in  std_logic
);
end entity;

architecture rtl of avalon_mm_basic_register_arbiter is

  signal address_i     : std_logic_vector(BASIC_REGISTER_ADDRESS_WIDTH - 1     downto 0);
  signal burst_count_i : std_logic_vector(BASIC_REGISTER_BURST_COUNT_WIDTH - 1 downto 0);
  signal read_i        : std_logic;
  signal write_i       : std_logic;
  signal write_data_i  : std_logic_vector(BASIC_REGISTER_DATA_WIDTH - 1        downto 0);

  signal interface_index  : unsigned(log_ceil(INPUTS) - 1 downto 0);
  signal pending_response : std_logic;

begin

  interface_sampling_proc : process(clk)
  begin
    if rising_edge(clk) then
      -- If performing a read hold onto this interface until a response has been detected
      if wait_request = '0' and read_i = '1' then
        pending_response <= '1';
      end if;

      -- If response happens clear hold
      if read_data_valid = '1' then
        pending_response <= '0';
      end if;

      -- Increment sampled interface index
      if (pending_response = '0' and not (wait_request = '0' and read = '1')) or read_data_valid = '1' then
        if to_integer(interface_index) < INPUTS - 1 then
          interface_index <= interface_index + 1;
        else
          interface_index <= (others => '0');
        end if;
      end if;

      if rst = '1' then
        pending_response <= '0';
        interface_index  <= (others => '0');
      end if;
    end if;
  end process;

  -- Selecting which input to sample (cleaner and allows us to check internal signals)
  address_i     <= address_array(to_integer(interface_index))     when rst = '0' else (others => '0');
  burst_count_i <= burst_count_array(to_integer(interface_index)) when rst = '0' else (others => '0');
  read_i        <= read_array(to_integer(interface_index))        when rst = '0' else '0';
  write_i       <= write_array(to_integer(interface_index))       when rst = '0' else '0';
  write_data_i  <= write_data_array(to_integer(interface_index))  when rst = '0' else (others => '0');

  -- passing the sampled interface through to the other side
  address_i     <= address;
  burst_count_i <= burst_count;
  read_i        <= read;
  write_i       <= write;
  write_data_i  <= write_data;

  -- Handling returning data
  response_handler_gen : for i in INPUTS - 1 downto 0 generate
    read_data_array       <= read_data                        when i = to_integer(interface_index) and rst = '0' else (others => '0');
    read_data_valid_array <= read_data_valid                  when i = to_integer(interface_index) and rst = '0' else '0';
    wait_request_array    <= wait_request or pending_response when i = to_integer(interface_index) and rst = '0' else '1';
  end generate;

end architecture;