-- This component is an interactable set of registers, which can listen to signals or set them in firmware
-- Intended to be used for debug in a HPS enabled system
-- It operates on a single clock
-- It should only listen to requests within it's address range
-- At the moment it can only operate with a burst count of 1
-- There must be at least 1 input register in this incarnation

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.math_pkg.all;
use work.avalon_mm_pkg.all;

entity interactable_basic_register is
generic
(
  BASE_ADDRESS     : std_logic_vector(31 downto 0) := (others => '0');
  BYTE_INDEXED     : boolean                       := false;
  INPUT_REGISTERS  : natural                       := 2;
  OUTPUT_REGISTERS : natural                       := 2
);
port
(
  clk : in  std_logic;
  rst : in  std_logic;

  address         : in  std_logic_vector(BASIC_REGISTER_ADDRESS_WIDTH - 1     downto 0);
  burst_count     : in  std_logic_vector(BASIC_REGISTER_BURST_COUNT_WIDTH - 1 downto 0);
  read            : in  std_logic;
  write           : in  std_logic;
  write_data      : in  std_logic_vector(BASIC_REGISTER_DATA_WIDTH - 1        downto 0);
  read_data       : out std_logic_vector(BASIC_REGISTER_DATA_WIDTH - 1        downto 0);
  read_data_valid : out std_logic;
  wait_request    : out std_logic;

  register_array_input  : in  BASIC_REGISTER_DATA_ARRAY((INPUT_REGISTERS + OUTPUT_REGISTERS) - 1 downto 0);
  register_array_output : out BASIC_REGISTER_DATA_ARRAY((INPUT_REGISTERS + OUTPUT_REGISTERS) - 1 downto 0)
);
end entity;

architecture rtl of interactable_basic_register is

  constant BYTES        : integer := 2 ** (log_ceil(BASIC_REGISTER_DATA_WIDTH) - 3);
  constant OUTPUT_START : integer := INPUT_REGISTERS;

  signal register_array_output_i : BASIC_REGISTER_DATA_ARRAY((INPUT_REGISTERS + OUTPUT_REGISTERS) - 1 downto 0)

begin

  wait_request <= not (read or write);

  registers_interaction_proc : process(clk)
    variable shifted_address : integer;
  begin
    if rising_edge(clk) then
      -- Applying the base address, then byte index if selected
      if BYTE_INDEXED = true then
        shifted_address := to_integer(unsigned(address)) / BYTES - to_integer(unsigned(BASE_ADDRESS)) / BYTES;
      else
        shifted_address := to_integer(unsigned(address)) - to_integer(unsigned(BASE_ADDRESS));
      end if;

      read_data_valid <= '0';

      -- Writes
      if write = '1' then
        if shifted_address >= OUTPUT_START and shifted_address < OUTPUT_START + OUTPUT_REGISTERS then
          register_array_output_i(shifted_address - OUTPUT_START) <= write_data;
        end if;
      end if;

      -- Reads
      if read = '1' then
        if shifted_address >= OUTPUT_START and shifted_address < OUTPUT_START + OUTPUT_REGISTERS then
          read_data       <= register_array_output_i(shifted_address - OUTPUT_START);
          read_data_valid <= '1';
        elsif shifted_address >= 0 and shifted_address < INPUT_REGISTERS then
          read_data       <= register_array_input(shifted_address);
          read_data_valid <= '1';
        end if;
      end if;

      if rst = '1' then
        read_data       <= (others => '0');
        read_data_valid <= '0';
      end if;
    end if;
  end process;

  register_array_output <= register_array_output_i;

end architecture;